# -*- coding: utf-8 -*-
import json
import os
from admix.helper import helper
import time
import shutil
from datetime import timezone, datetime, timedelta

from admix.interfaces.database import ConnectMongoDB
from admix.helper.decorator import Collector

#get Rucio imports done:
from admix.interfaces.rucio_dataformat import ConfigRucioDataFormat
from admix.interfaces.rucio_summoner import RucioSummoner
from admix.interfaces.destination import Destination
from admix.interfaces.keyword import Keyword
from admix.interfaces.templater import Templater
from admix.utils import make_did
from admix.utils.list_file_replicas import list_file_replicas

@Collector
class CleanEB():

    def __init__(self):
        pass

    def init(self):
        helper.global_dictionary['logger'].Info(f'Init task {self.__class__.__name__}')


        #Define data types
        self.NORECORDS_DTYPES = helper.get_hostconfig()['norecords_types']
        self.RAW_RECORDS_DTYPES = helper.get_hostconfig()['raw_records_types']
        self.RECORDS_DTYPES = helper.get_hostconfig()['records_types']

        #Get other parameters
        self.DATADIR = helper.get_hostconfig()['path_data_to_upload']
        self.RUCIODATADIR = helper.get_hostconfig()['path_rucio_data']
        self.periodic_check = helper.get_hostconfig()['upload_periodic_check']
        self.RSES = helper.get_hostconfig()['rses']

        # Choose which RSE is used for the upload (usually it is LNGS_USERDISK)
        self.UPLOAD_TO = helper.get_hostconfig()['upload_to']

        self.minimum_number_acceptable_rses = 1
        self.minimum_deltadays_allowed = 0 #3
        self.minimum_deltadays_allowed_heavy = 0 #1
        self.dtype_delayed_delete = ['raw_records_aqmon','raw_records_aqmon_nv','raw_records_he','raw_records_mv','raw_records_nv','pulse_counts','pulse_counts_he','veto_regions','peaklets','peaklets_he','records_he']
        self.dtype_delayed_delete_heavy = ['raw_records','records']
        self.dtype_never_delete = ['lone_hits','merged_s2s','peak_basics','peaklet_classification']

        #Init the runDB
        self.db = ConnectMongoDB()

        #We want the first and the last run:
        self.gboundary = self.db.GetBoundary()
        self.run_nb_min = self.gboundary['min_number']
        self.run_nb_max = self.gboundary['max_number']
        self.run_ts_min = self.gboundary['min_start_time']
        self.run_ts_max = self.gboundary['max_start_time']

        #Init the Rucio data format evaluator in three steps:
        self.rc_reader = ConfigRucioDataFormat()
        self.rc_reader.Config(helper.get_hostconfig('rucio_template'))

        #This class will evaluate your destinations:
        self.destination = Destination()

        #Since we deal with an experiment, everything is predefine:
        self.exp_temp = Templater()
        self.exp_temp.Config(helper.get_hostconfig()['template'])

        #Init a class to handle keyword strings:
        self.keyw = Keyword()

        #Init Rucio for later uploads and handling:
        self.rc = RucioSummoner(helper.get_hostconfig("rucio_backend"))
        self.rc.SetRucioAccount(helper.get_hostconfig('rucio_account'))
        self.rc.SetConfigPath(helper.get_hostconfig("rucio_cli"))
        self.rc.SetProxyTicket(helper.get_hostconfig('rucio_x509'))
        self.rc.SetHost(helper.get_hostconfig('host'))
        self.rc.ConfigHost()
        self.rc.SetProxyTicket("rucio_x509")
#        print(self.rc.Whoami())





    def run(self,*args, **kwargs):
        helper.global_dictionary['logger'].Info(f'Run task {self.__class__.__name__}')


        data_types = self.RAW_RECORDS_DTYPES + self.RECORDS_DTYPES + self.NORECORDS_DTYPES
#        data_types = self.RAW_RECORDS_DTYPES + self.RECORDS_DTYPES


        # Get all runs that are already transferred and that still have some data_types in eb 
        cursor = self.db.db.find({
#            'number': {"$lt": 7330, "$gte": 7300},
#            'number': {"$lt": 8570, "$gte": 8550},
#            'number': {"$gte": 7330},
            'number': {"$gte": 8500},
#            'number': 9226,
#            'data' : { "$elemMatch": { "host" : {"$regex" : ".*eb.*"} , "type" : {"$in" : data_types}} },
            'status': 'transferred'
        },
        {'_id': 1, 'number': 1, 'data': 1, 'bootstrax': 1})


        cursor = list(cursor)

        helper.global_dictionary['logger'].Info('Runs that will be processed are {0}'.format([c["number"] for c in cursor]))

        # Runs over all listed runs
        for run in cursor:

            #Gets the run number
            number = run['number']

            # Extracts the correct Event Builder machine who processed this run
            bootstrax = run['bootstrax']
            eb = bootstrax['host'].split('.')[0]

            helper.global_dictionary['logger'].Info('Treating run {0}'.format(number))

            helper.global_dictionary['logger'].Info('Run {0} has been processed by {1}'.format(number,eb))

            # Checks how much date are old
            run_time = run['bootstrax']['time'].replace(tzinfo=timezone.utc)
            now_time = datetime.now().replace(tzinfo=timezone.utc)
            delta_time = now_time - run_time

            # Loops on all datatypes that have to be cleaned
            for dtype in data_types:
                helper.global_dictionary['logger'].Info('\t==> Looking for data type {0}'.format(dtype))

                # checks the age of the data type
                is_enough_old = True

                # for some data types, it they are not yet older than three days, it skips deleting them
                if dtype in self.dtype_delayed_delete:
                    if delta_time < timedelta(days=self.minimum_deltadays_allowed):
                        helper.global_dictionary['logger'].Info('Run {0}, data type {1} is not yet older than {2} days. Skip it'.format(number,dtype,self.minimum_deltadays_allowed))
                        is_enough_old = False

                # for some heavy data types (records and raw_records), if they are not yet older than one day, it skips deleting them
                if dtype in self.dtype_delayed_delete_heavy:
                    if delta_time < timedelta(days=self.minimum_deltadays_allowed_heavy):
                        helper.global_dictionary['logger'].Info('Run {0}, data type {1} is not yet older than {2} days. Skip it'.format(number,dtype,self.minimum_deltadays_allowed_heavy))
                        is_enough_old = False

                # check first with runDB if the data type already exists in external RSEs
                rses_in_db = []
                for d in run['data']:
                    if d['type'] == dtype and d['host'] == 'rucio-catalogue' and d['location'] != self.UPLOAD_TO:
                        rses_in_db.append(d['location'])
                helper.global_dictionary['logger'].Info('\t==> According to DB, found in following external RSEs : {0}'.format(rses_in_db))

                # if this is not the case, just skip any attempt of deleting anything
                if len(rses_in_db) < self.minimum_number_acceptable_rses:
                    helper.global_dictionary['logger'].Info('\t==> Nothing will be deleted : not enough external RSEs')
                    continue

                #
                # Phase 1 : Deleting data in EB
                #

                # check first if data are, according to the DB, still in EB
                datum = None
                for d in run['data']:
                    if d['type'] == dtype and eb in d['host']:
                        datum = d

                if datum is None:
                    helper.global_dictionary['logger'].Info('Data type not in eb')

                # start deleting data in EB
                if datum is not None and dtype not in self.dtype_never_delete and is_enough_old:
                    file = datum['location'].split('/')[-1]
                    hash = file.split('-')[-1]

                    # create the DID from DB
                    did = make_did(number, dtype, hash)

                    # check if a rule already exists with this exact DID in external RSEs
                    # and take also the number of files in each RSE
                    rses_with_rule = []
                    rses_with_correct_nfiles = []
                    for rse in self.RSES:
                        rucio_rule = self.rc.GetRule(upload_structure=did, rse=rse)
                        if rucio_rule['exists'] and rucio_rule['state'] == 'OK':
                            if self.UPLOAD_TO==rucio_rule['rse']:
                                continue
                            rses_with_rule.append(rucio_rule['rse'])
                            nfiles = len(list_file_replicas(number, dtype, hash, rucio_rule['rse']))
                            if 'file_count' in datum:
                                if nfiles == datum['file_count']:
                                    rses_with_correct_nfiles.append(rucio_rule['rse'])
                    helper.global_dictionary['logger'].Info('\t==> According to Rucio, found in following external RSEs : {0}'.format(rses_with_rule))

                    if len(rses_with_correct_nfiles) == len(rses_with_rule):
                        helper.global_dictionary['logger'].Info('\t==> All of them with the expected number of files')
                    else:
                        helper.global_dictionary['logger'].Info('\t==> Error, these RSEs have wrong number of files : {0}'.format(rses_with_correct_nfiles))
                    
                    # if so, start deleting
#                    if len(rses_with_rule)>=self.minimum_number_acceptable_rses and len(rses_with_correct_nfiles) == len(rses_with_rule):
                    if len(rses_with_rule)>=self.minimum_number_acceptable_rses:

                        # delete from DB
                        # print(run['_id'],datum['type'],datum['host'])
                        self.db.RemoveDatafield(run['_id'],datum)
                        full_path = os.path.join(self.DATADIR, eb, file)
                        # print(full_path)

                        helper.global_dictionary['logger'].Info('\t==> Deleted EB info from DB') 

                        # delete from disk
                        try:
                            shutil.rmtree(full_path)
                        except OSError as e:
                            helper.global_dictionary['logger'].Info('\t==> Error, cannot delete directory : {0}'.format(e))
                        else:
                            helper.global_dictionary['logger'].Info('\t==> Deleted data from EB disk') 


                #
                # Phase 2 : Deleting data in LNGS_USERDISK
                #

                # check if data are, according to the DB, still in datamanager (LNGS_USERDISK)
                datum = None
                for d in run['data']:
                    if d['type'] == dtype and d['host'] == 'rucio-catalogue' and self.UPLOAD_TO in d['location']:
                        datum = d

                # if so, start deleting data in datamanager (LNGS_USERDISK)
                if datum is None:
                    helper.global_dictionary['logger'].Info('Data type not in LNGS_USERDISK')
                else:
                    # create the DID from DB
                    did = datum['did']
                    hash = did.split('-')[-1]

                    nfiles_upload_to = len(list_file_replicas(number, dtype, hash, self.UPLOAD_TO))

                    # check if a rule already exists with this exact DID in external RSEs
                    rses_with_rule = []
                    rses_with_correct_nfiles = []
                    for rse in self.RSES:
                        rucio_rule = self.rc.GetRule(upload_structure=did, rse=rse)
                        if rucio_rule['exists'] and rucio_rule['state'] == 'OK':
                            if self.UPLOAD_TO==rucio_rule['rse']:
                                continue
                            rses_with_rule.append(rucio_rule['rse'])
                            nfiles = len(list_file_replicas(number, dtype, hash, rucio_rule['rse']))
                            if nfiles == nfiles_upload_to:
                                rses_with_correct_nfiles.append(rucio_rule['rse'])
                    helper.global_dictionary['logger'].Info('\t==> According to Rucio, found in following external RSEs : {0}'.format(rses_with_rule))

                    if len(rses_with_correct_nfiles) == len(rses_with_rule):
                        helper.global_dictionary['logger'].Info('\t==> All of them with the expected number of files')
                    else:
                        helper.global_dictionary['logger'].Info('\t==> Error, these RSEs have wrong number of files : {0}'.format(rses_with_correct_nfiles))

                    # if so, start deleting
                    if len(rses_with_rule)>=self.minimum_number_acceptable_rses and len(rses_with_correct_nfiles) == len(rses_with_rule):

                        rucio_rule = self.rc.GetRule(upload_structure=did, rse=self.UPLOAD_TO)
                        if rucio_rule['exists'] and rucio_rule['state'] == 'OK' and rucio_rule['rse'] == self.UPLOAD_TO:
                            self.rc.DeleteRule(rucio_rule['id'])
                            helper.global_dictionary['logger'].Info('\t==> Deleted LNGS_USERDISK Rucio rule') 
                            hash = did.split('-')[-1]
                            files = list_file_replicas(number, dtype, hash, "LNGS_USERDISK")                        
                            for file in files:
                                os.remove(file)
                            helper.global_dictionary['logger'].Info('\t==> Deleted data from LNGS_USERDISK disk') 

                            self.db.RemoveDatafield(run['_id'],datum)
                            helper.global_dictionary['logger'].Info('\t==> Deleted LNGS_USERDISK info from DB') 
                    

        return 0


    def __del__(self):
        pass



"""
.. module:: rucio_cli
   :platform: Unix
   :synopsis: A legacy wrapper for the Rucio command line interface (CLI)

.. moduleauthor:: Boris Bauermeister <Boris.Bauermeister@gmail.com>

"""

import logging
import tempfile
import subprocess
import datetime
import os
from admix.helper.decorator import Collector

@Collector
class RucioCLI():

    def __init__(self):
        self.rucio_version = None
        self.rucio_account = None
        self.rucio_host = None

    # Here comes the backend configuration part:
    def SetConfigPath(self, config_path):
        self.rucio_cli_config = config_path

    def SetHost(self, host):
        self.rucio_host = host

    def SetProxyTicket(self, ticket_path):
        self.rucio_ticket_path = ticket_path

    def SetRucioAccount(self, account):
        self.rucio_account = account

    def ConfigHost(self):
        try:
            self.config = self.GetConfig().format(rucio_account=self.rucio_account, x509_user_proxy=self.rucio_ticket_path)
        except:
            print("Can not init the Rucio CLI")
            print("-> Check for your Rucio installation")
            exit(1)

    def Alive(self):
        print("RucioCLI alive")
    #Finished the backend configuration for RucioCLI

    #Run Bash scripts out of Python:
    def create_script(self, script):
        """Create script as temp file to be run on cluster"""
        fileobj = tempfile.NamedTemporaryFile(delete=False,
                                            suffix='.sh',
                                            mode='wt',
                                            #bufsize=1)
                                            buffering=1)
        fileobj.write(script)
        os.chmod(fileobj.name, 0o774)
        return fileobj

    def delete_script(self, fileobj):
        """Delete script after submitting to cluster
        :param script_path: path to the script to be removed
        """
        fileobj.close()

    def doRucio(self, upload_string ):
        sc = self.create_script( upload_string )
        execute = subprocess.Popen( ['sh', sc.name] ,
                                    stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    shell=False,
                                    universal_newlines=False)
        stdout_value, stderr_value = execute.communicate()
        stdout_value = stdout_value.decode("utf-8")
        stdout_value = stdout_value.split("\n")
        stdout_value = list(filter(None, stdout_value)) # fastest way to remove '' from list
        self.delete_script(sc)
        return stdout_value, stderr_value


    def GetConfig(self):
        config_string = ""

        if os.path.isdir(self.rucio_cli_config) and self.rucio_host != None:
            f_file = open(os.path.join(self.rucio_cli_config, 'rucio_cli_{host}.config'.format(host=self.rucio_host)),"r")
            for i_line in f_file:
                #i_line=i_line.replace("\n", "")
                if i_line.find("#!/7")==0:
                    continue
                config_string+=i_line
        else:

            print("Log: Miss config")

        return config_string



    def Whoami(self):
        """RucioCLI:Whoami
        Results a dictionary to identify the current
        Rucio user and credentials.
        CLI call requires to parse return string (msg)
        into a dictionary first.
        """
        upload_string = "rucio whoami"
        upload_string = self.config + upload_string
        msg, err = self.doRucio(upload_string)

        dict_msg = {}
        for i_msg in msg:
            i_msg = i_msg.replace(" ", "")
            key, value = i_msg.split(":", 1)
            dict_msg[key] = value
        return dict_msg

    def ListDidRules(self, scope, dname):
        """List rules by scope and name from the command line
        call.
        """

        pass
        #upload_string = f"rucio list-rules {scope}:{dname}"
        #upload_string = self.config + upload_string
        #msg, err = self.doRucio(upload_string)

        dict_msg = {}
        ...


    def CliUpload(self, method=None, upload_dict={}):

        upload_string = ""
        if method == "upload-folder-with-did":
            upload_string = "rucio upload --rse {rse} --scope {scope} {scope}:{did} {datasetpath}\n".format(rse=upload_dict['rse'],scope=upload_dict['scope'],did=upload_dict['did'], datasetpath=upload_dict['upload_path'])

        #cliupload:
        upload_string = self.config + upload_string

        print(upload_string)
        msg, err = self.doRucio(upload_string)

        status = 'OK'
        status_msg = []
        for i_msg in msg:
            if i_msg.find("ERROR")>= 0:
                status = "ERROR"
            status_msg.append(i_msg)

        return status, status_msg



# -*- coding: utf-8 -*-
import json
import os
from admix.helper import helper
import time
import shutil

from admix.interfaces.database import ConnectMongoDB
from admix.helper.decorator import Collector

#get Rucio imports done:
from admix.interfaces.rucio_dataformat import ConfigRucioDataFormat
from admix.interfaces.rucio_summoner import RucioSummoner
from admix.interfaces.destination import Destination
from admix.interfaces.keyword import Keyword
from admix.interfaces.templater import Templater
from admix.utils import make_did

@Collector
class UploadFromLNGS():

    def __init__(self):
        pass

    def init(self):
        helper.global_dictionary['logger'].Info(f'Init task {self.__class__.__name__}')


#        open("/tmp/admix-upload_from_lngs", 'a').close()

        #Define data types
        self.DTYPES = helper.get_hostconfig()['rawtype']
#        self.DTYPES = ["pulse_counts"]
        self.DATADIR = helper.get_hostconfig()['path_data_to_upload']
        self.periodic_check = helper.get_hostconfig()['upload_periodic_check']

        #Init the runDB
        self.db = ConnectMongoDB()

        #We want the first and the last run:
        self.gboundary = self.db.GetBoundary()
        self.run_nb_min = self.gboundary['min_number']
        self.run_nb_max = self.gboundary['max_number']
        self.run_ts_min = self.gboundary['min_start_time']
        self.run_ts_max = self.gboundary['max_start_time']

        #Init the Rucio data format evaluator in three steps:
        self.rc_reader = ConfigRucioDataFormat()
        self.rc_reader.Config(helper.get_hostconfig('rucio_template'))

        #This class will evaluate your destinations:
        self.destination = Destination()

        #Since we deal with an experiment, everything is predefine:
        self.exp_temp = Templater()
        self.exp_temp.Config(helper.get_hostconfig()['template'])

        #Init a class to handle keyword strings:
        self.keyw = Keyword()

        #Init Rucio for later uploads and handling:
        self.rc = RucioSummoner(helper.get_hostconfig("rucio_backend"))
        self.rc.SetRucioAccount(helper.get_hostconfig('rucio_account'))
        self.rc.SetConfigPath(helper.get_hostconfig("rucio_cli"))
        self.rc.SetProxyTicket(helper.get_hostconfig('rucio_x509'))
        self.rc.SetHost(helper.get_hostconfig('host'))
        self.rc.ConfigHost()
        self.rc.SetProxyTicket("rucio_x509")
#        print(self.rc.Whoami())


    def find_data_to_upload(self):
#        cursor = self.db.db.find({'status': 'needs_upload'}, {'number': 1, 'data': 1})
        cursor = self.db.db.find({'status': 'eb_ready_to_upload'}, {'number': 1, 'data': 1})
        ids = []

        for r in cursor:
            dtypes = set([d['type'] for d in r['data']])
            # check if all of the necessary data types are in the database
            if set(self.DTYPES) <= dtypes:
                ids.append(r['_id'])
        return ids




    def check_transfers(self):
        cursor = self.db.db.find(
            {'status': 'transferring'},
#            {'number':7185},
            {'number': 1, 'data': 1})

        cursor = list(cursor)

        helper.global_dictionary['logger'].Info('Checking transfer status of {0} runs'.format(len(cursor)))

        for run in list(cursor):
            # for each run, check the status of all REPLICATING rules
            rucio_stati = []
            for d in run['data']:
                if d['host'] == 'rucio-catalogue':
#                    if run['number']==7695 and d['status'] == 'error':
#                        self.db.db.find_one_and_update({'_id': run['_id'],'data': {'$elemMatch': d}},
#                                                       {'$set': {'data.$.status': 'transferring'}}
#                                                   )                        
                    if d['status'] != 'transferring':
                        rucio_stati.append(d['status'])
                    else:
                        did = d['did']
                        status = self.rc.CheckRule(did, d['location'])
                        if status == 'REPLICATING':
                            rucio_stati.append('transferring')
                        elif status == 'OK':
                            # update database
                            helper.global_dictionary['logger'].Info('Updating DB for run {0}, dtype {1}, location {2}'.format(run['number'], d['type'],d['location']))
                            self.db.db.find_one_and_update({'_id': run['_id'],'data': {'$elemMatch': d}},
                                                      {'$set': {'data.$.status': 'transferred'}}
                            )
                            rucio_stati.append('transferred')

                        elif status == 'STUCK':
                            self.db.db.find_one_and_update({'_id': run['_id'], 'data': {'$elemMatch': d}},
                                                      {'$set': {'data.$.status': 'error'}}
                            )
                            rucio_stati.append('error')

                            # are there any other rucio rules transferring?
            if len(rucio_stati) > 0 and all([s == 'transferred' for s in rucio_stati]):
                self.db.SetStatus(run['number'], 'transferred')



    def add_rule(self,run_number, dtype, hash, rse, lifetime=None, update_db=True):
        did = make_did(run_number, dtype, hash)
        result = self.rc.AddRule(did, rse, lifetime=lifetime)
        #if result == 1:
        #   return
        helper.global_dictionary['logger'].Info('Rule Added: {0} ---> {1}'.format(did,rse))

        if update_db:
            rucio_rule = self.rc.GetRule(did, rse=rse)
            data_dict = {'host': "rucio-catalogue",
                         'type': dtype,
                         'location': rse,
                         'lifetime': rucio_rule['expires'],
                         'status': 'transferring',
                         'did': did,
                         'protocol': 'rucio'
                     }
            self.db.db.find_one_and_update({'number': run_number},
                                      {'$set': {'status': 'transferring'}}
                                  )

            docid = self.db.db.find_one({'number': run_number}, {'_id': 1})['_id']
            self.db.AddDatafield(docid, data_dict)



    def remove_from_eb(number, dtype):
        query = {'number': number}
        cursor = self.db.db.find_one(query, {'number': 1, 'data': 1})
        ebdict = None
        for d in cursor['data']:
            if '.xenon.local' in d['host'] and d['type'] == dtype:
                ebdict = d

        # get name of file (really directory) of this dtype in eb storage
        if ebdict is None:
            helper.global_dictionary['logger'].Info('No eventbuilder datum found for run {0} {1} Exiting.'.format(number,dtype))
            return

        file = ebdict['location'].split('/')[-1]
        path_to_rm = os.path.join(self.DATADIR, file)

        helper.global_dictionary['logger'].Info('{0}'.format(path_to_rm))
        helper.global_dictionary['logger'].Info('{0}'.format(ebdict))
        shutil.rmtree(path_to_rm)
        self.db.RemoveDatafield(cursor['_id'], ebdict)




    def run(self,*args, **kwargs):
        helper.global_dictionary['logger'].Info(f'Run task {self.__class__.__name__}')


        # Finds new data
#        query = {'number': {'$gte': 7158},
#        query = {'number': {'$gte': 7150},
#                 "status": {"$exists": False},
#             }

#        cursor = self.db.db.find(query, {'number': 1})

#        for r in cursor:
#            self.db.SetStatus(r['number'], 'needs_upload')


        # List runs to upload
        ids_to_upload = self.find_data_to_upload()

        cursor = self.db.db.find({'_id': {"$in": ids_to_upload},
                             'number': 7947
#                             'number': 7177
#                                  'number': {"$gte": 7807}
                         },
                            {'number': 1, 'data': 1})

        cursor = list(cursor)

        helper.global_dictionary['logger'].Info('Runs that will be processed are {0}'.format([c["number"] for c in cursor]))

        # Check transfers
        self.check_transfers()
        last_check = time.time()

        # Performs uploads to all listed runs
        for run in cursor:
            number = run['number']
            helper.global_dictionary['logger'].Info('Uploading run {0}'.format(number))
            for dtype in self.DTYPES:
                helper.global_dictionary['logger'].Info('\t==> Uploading {0}'.format(dtype))
                # get the datum for this datatype
                datum = None
                in_rucio = False
                for d in run['data']:
                    if d['type'] == dtype and 'eb' in d['host']:
                        datum = d

                    if d['type'] == dtype and d['host'] == 'rucio-catalogue':
                        in_rucio = True

                if datum is None:
                    helper.global_dictionary['logger'].Info('Data type {0} not found for run {1}'.format(dtype,number))
                    continue

                file = datum['location'].split('/')[-1]

                hash = file.split('-')[-1]

                upload_path = os.path.join(self.DATADIR, file)

                # create a DID to upload
                did = make_did(number, dtype, hash)

                # check if a rule already exists for this DID on LNGS
                rucio_rule = self.rc.GetRule(upload_structure=did, rse="LNGS_USERDISK")
                helper.global_dictionary['logger'].Info('It was already in Rucio : {0}'.format(in_rucio))
                helper.global_dictionary['logger'].Info('Rucio rule : {0}'.format(rucio_rule))

                # if not in rucio already and no rule exists, upload into rucio
                if not in_rucio and not rucio_rule['exists']:
                    result = self.rc.Upload(did,
                                       upload_path,
                                       'LNGS_USERDISK',
                                       lifetime=None)
                    helper.global_dictionary['logger'].Info('Data type uploaded.')

                # if upload was successful, tell runDB
                rucio_rule = self.rc.GetRule(upload_structure=did, rse="LNGS_USERDISK")
                data_dict = {'host': "rucio-catalogue",
                             'type': dtype,
                             'location': 'LNGS_USERDISK',
                             'lifetime': rucio_rule['expires'],
                             'status': 'transferred',
                             'did': did,
                             'protocol': 'rucio'
                         }

                if rucio_rule['state'] == 'OK':
                    if not in_rucio:
                        self.db.AddDatafield(run['_id'], data_dict)

                    # add a DID list that's easy to query by DB.GetDid
                    # check if did field exists yet or not
#                    if not run.get('dids'):
#                        self.db.db.find_one_and_update({'_id': run['_id']},
#                                                  {'$set': {'dids': {dtype: did}}}
#                        )
#                    else:
#                        helper.global_dictionary['logger'].Info('Updating DID list')
#                        self.db.db.find_one_and_update({'_id': run['_id']},
#                                                  {'$set': {'dids.%s' % dtype: did}}
#                        )

                # add rule to OSG and Nikhef
                # TODO make this configurable
#                for rse in ['UC_OSG_USERDISK', 'UC_DALI_USERDISK']:
#                for rse in ['UC_OSG_USERDISK', 'UC_DALI_USERDISK','CNAF_TAPE2_USERDISK','CNAF_USERDISK','NIKHEF2_USERDISK','CCIN2P3_USERDISK']:
#                for rse in ['UC_OSG_USERDISK', 'UC_DALI_USERDISK','NIKHEF2_USERDISK']:
#                for rse in ['UC_DALI_USERDISK','NIKHEF2_USERDISK']:
#                for rse in ['UC_OSG_USERDISK','NIKHEF2_USERDISK']:
                for rse in ['UC_OSG_USERDISK']:
#                for rse in ['UC_OSG_USERDISK','CCIN2P3_USERDISK']:
                    self.add_rule(number, dtype, hash, rse)

#                if number==7558:
#                    self.add_rule(number, dtype, hash, 'CNAF_TAPE2_USERDISK')

                # finally, delete the eb copy
                #self.remove_from_eb(number, dtype)

            if time.time() - last_check > self.periodic_check:
                self.check_transfers()
                last_check = time.time()


        return 0


    def __del__(self):
        pass



#Here we have our decorators:

NameCollector = []
ClassCollector = {}

class Collector():
    def __init__(self, handover):
        if handover.__name__ not in NameCollector:
            NameCollector.append(handover.__name__)
            ClassCollector[str(handover.__name__)] = handover.__call__()

    def __call__(self, *args, **kwargs):
        pass



import os
import sys
from argparse import ArgumentParser
from admix.interfaces.rucio_summoner import RucioSummoner
from admix.interfaces.database import ConnectMongoDB
from admix.utils.naming import make_did
try:
    from straxen import __version__
    straxen_version = __version__
except ImportError:
    print("Straxen not installed in current env, so must pass straxen_version manually")
import time
import utilix

DB = ConnectMongoDB()

def determine_rse(rse_list, glidein_country):
    # TODO put this in config or something?
    EURO_SITES = ["CCIN2P3_USERDISK",
                  "NIKHEF_USERDISK",
                  "NIKHEF2_USERDISK",
                  "WEIZMANN_USERDISK",
                  "CNAF_USERDISK",
                  "SURFSARA_USERDISK"]

    US_SITES = ["UC_OSG_USERDISK", "UC_DALI_USERDISK"]


    if glidein_country == "US":
        for site in US_SITES:
            if site in rse_list:
                return site

    elif glidein_country == "FR":
        for site in EURO_SITES:
            if site in rse_list:
                return site

    elif glidein_country == "NL":
        for site in reversed(EURO_SITES):
            if site in rse_list:
                return site

    elif glidein_country == "IL":
        for site in EURO_SITES:
            if site in rse_list:
                return site

    elif glidein_country == "IT":
        for site in EURO_SITES:
            if site in rse_list:
                return site

    if US_SITES[0] in rse_list:
        return US_SITES[0]
    else:
        raise AttributeError("cannot download data")


def download(number, dtype, hash, chunks=None, location='.',  tries=3, metadata=True,
             num_threads=3, **kwargs):
    """Function download()

    Downloads a given run number using rucio
    :param number: A run number (integer)
    :param dtype: The datatype to download.
    :param chunks: List of integers representing the desired chunks. If None, the whole run will be downloaded.
    :param location: String for the path where you want to put the data. Defaults to current directory.
    :param tries: Integer specifying number of times to try downloading the data. Defaults to 2.
    :param version: Context version as listed in the data_hashes collection
    :param kwargs: Keyword args passed to DownloadDids
    """

    # setup rucio client
    rc = RucioSummoner()

    # get DID
    did = make_did(number, dtype, hash)

    # if we didn't pass an rse, determine the best one
    rse = kwargs.pop('rse', None)
    if not rse:
        # determine which rses this did is on
        rules = rc.ListDidRules(did)
        rses = []
        for r in rules:
            if r['state'] == 'OK':
                rses.append(r['rse_expression'])
        # find closest one
        rse = determine_rse(rses, os.environ.get('GLIDEIN_Country', 'US'))

    if chunks:
        dids = []
        for c in chunks:
            cdid = did + '-' + str(c).zfill(6)
            dids.append(cdid)
        # also download metadata
        if metadata:
            dids.append(did + '-metadata.json')

    else:
        dids = [did]

    # rename the folder that will be downloaded
    path = did.replace(':', '-')
    # drop the xnt at the beginning
    path = path.replace('xnt_', '')

    location = os.path.join(location, path)
    os.makedirs(location, exist_ok=True)

    # TODO check if files already exist?

    print(f"Downloading {did} from {rse}")

    _try = 1
    success = False

    while _try <= tries and not success:
        if _try == tries:
            rse = None
        result = rc.DownloadDids(dids, download_path=location, no_subdir=True, rse=rse,
                                 num_threads=num_threads, **kwargs)
        if isinstance(result, int):
            print(f"Download try #{_try} failed.")
            _try += 1
            time.sleep(5)
        else:
            success = True

    if success:
        print(f"Download successful to {location}")


def main():
    parser = ArgumentParser("admix-download")

    parser.add_argument("number", type=int, help="Run number to download")
    parser.add_argument("dtype", help="Data type to download")
    parser.add_argument("--chunks", nargs="*", help="Space-separated list of chunks to download.")
    parser.add_argument("--dir", help="Path to put the downloaded data.", default='.')
    parser.add_argument('--tries', type=int, help="Number of tries to download the data.", default=2)
    parser.add_argument('--rse', help='RSE to download from')
    parser.add_argument('--threads', help='Number of threads to use', default=3, type=int)
    parser.add_argument('--context', help='strax context you need -- this determines the hash',
                         default='xenonnt_online')
    parser.add_argument('--straxen_version', help='straxen version', default=None)

    args = parser.parse_args()

    # use system straxen version if none passed
    version = args.straxen_version if args.straxen_version else straxen_version
    hash = utilix.db.get_hash(args.context, args.dtype, version)
    if args.chunks:
        chunks = [int(c) for c in args.chunks]
    else:
        chunks=None

    download(args.number, args.dtype, hash, chunks=chunks, location=args.dir, tries=args.tries,
             rse=args.rse, num_threads=args.threads)

